﻿// ManipulateCharacterOutlines (c) 2009 by Charles Petzold
using System;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Collections.ObjectModel;

namespace ManipulateCharacterOutlines
{
    public partial class MainPage : UserControl
    {
        const double THUMB_SIZE = 8;
        const double MIN_WIDTH_HEIGHT = 12;

        Thumb activeBackgroundThumb;
        Thumb[] cornerThumbs = new Thumb[4];
        Collection<Thumb> nodeThumbs = new Collection<Thumb>();

        public MainPage()
        {
            InitializeComponent();

            double x = 0;
            Brush blueBrush = new SolidColorBrush(Colors.Blue);

            for (int i = 0; i < pathGeometryCollection.Items.Count; i++)
            {
                // Create letter from PathGeometry
                Path path = new Path()
                {
                    Fill = blueBrush,
                    Data = pathGeometryCollection.Items[i] as PathGeometry,
                    IsHitTestVisible = false
                };

                // Create Thumb objects for background to letters
                Thumb thumb = new Thumb()
                {
                    Width = letterSizes.Points[i].X,
                    Height = letterSizes.Points[i].Y,
                    Opacity = 0,
                    Tag = path.Data
                };
                thumb.DragStarted += new DragStartedEventHandler(OnLetterMoveThumbDragStarted);
                thumb.DragDelta += new DragDeltaEventHandler(OnBackgroundThumbDragDelta);

                Canvas.SetLeft(thumb, x);
                Canvas.SetTop(thumb, 0);
                characterOutlineCanvas.Children.Add(thumb);
                x += letterSizes.Points[i].X;

                characterOutlineCanvas.Children.Add(path);
            }
        }

        void OnLetterMoveThumbDragStarted(object sender, DragStartedEventArgs args)
        {
            Thumb backgroundThumb = sender as Thumb;

            if (backgroundThumb == activeBackgroundThumb)
                return;

            CleanUp();

            // Create rectangle corner thumbs
            for (int corner = 0; corner < 4; corner++)
            {
                CornerThumbTargetInfo cornerThumbTarget = new CornerThumbTargetInfo()
                {
                    BackgroundThumb = backgroundThumb,
                    Corner = (Corner)corner
                };

                // Create corner thumbs
                Thumb cornerThumb = new Thumb()
                {
                    Width = THUMB_SIZE,
                    Height = THUMB_SIZE,
                    Tag = cornerThumbTarget
                };
                cornerThumb.DragDelta += new DragDeltaEventHandler(OnCornerThumbDragDelta);

                // Position thumbs on four corners
                switch ((Corner)corner)
                {
                    case Corner.UL:
                        Canvas.SetLeft(cornerThumb, Canvas.GetLeft(backgroundThumb) - cornerThumb.Width / 2);
                        Canvas.SetTop(cornerThumb, Canvas.GetTop(backgroundThumb) - cornerThumb.Height / 2);
                        break;

                    case Corner.UR:
                        Canvas.SetLeft(cornerThumb, Canvas.GetLeft(backgroundThumb) - cornerThumb.Width / 2 + backgroundThumb.Width);
                        Canvas.SetTop(cornerThumb, Canvas.GetTop(backgroundThumb) - cornerThumb.Height / 2);
                        break;

                    case Corner.LL:
                        Canvas.SetLeft(cornerThumb, Canvas.GetLeft(backgroundThumb) - cornerThumb.Width / 2);
                        Canvas.SetTop(cornerThumb, Canvas.GetTop(backgroundThumb) - cornerThumb.Height / 2 + backgroundThumb.Height);
                        break;

                    case Corner.LR:
                        Canvas.SetLeft(cornerThumb, Canvas.GetLeft(backgroundThumb) - cornerThumb.Width / 2 + backgroundThumb.Width);
                        Canvas.SetTop(cornerThumb, Canvas.GetTop(backgroundThumb) - cornerThumb.Height / 2 + backgroundThumb.Height);
                        break;
                }

                characterOutlineCanvas.Children.Add(cornerThumb);
                cornerThumbs[corner] = cornerThumb;
            }

            // Create node thumbs
            PathGeometry pathGeometry = backgroundThumb.Tag as PathGeometry;
            nodeThumbs.Clear();

            PathGeometryHelper.Enumerate(pathGeometry, 
                delegate(object targetObject, PropertyInfo propertyInfo)
                {
                    if (propertyInfo.PropertyType == typeof(Point))
                    {
                        NodeThumbTargetInfo nodeThumbTarget = new NodeThumbTargetInfo()
                        {
                            BackgroundThumb = backgroundThumb,
                            TargetObject = targetObject,
                            PropertyInfo = propertyInfo
                        };

                        Thumb nodeThumb = new Thumb()
                        {
                            Width = THUMB_SIZE,
                            Height = THUMB_SIZE,
                            Tag = nodeThumbTarget
                        };
                        nodeThumb.DragDelta += OnNodeThumbDragDelta;

                        ToolTipService.SetToolTip(nodeThumb, 
                            String.Format("{0}.{1}", targetObject.GetType().Name, 
                                                     propertyInfo.Name));

                        Point pt = (Point)propertyInfo.GetValue(targetObject, null);
                        Canvas.SetLeft(nodeThumb, pt.X - nodeThumb.Width / 2);
                        Canvas.SetTop(nodeThumb, pt.Y - nodeThumb.Height / 2);

                        characterOutlineCanvas.Children.Add(nodeThumb);
                        nodeThumbs.Add(nodeThumb);
                    }
                    else if (propertyInfo.PropertyType == typeof(PointCollection))
                    {
                        PointCollection ptColl = propertyInfo.GetValue(targetObject, null) as PointCollection;

                        for (int index = 0; index < ptColl.Count; index++)
                        {
                            NodeThumbTargetInfo nodeThumbTarget = new NodeThumbTargetInfo()
                            {
                                BackgroundThumb = backgroundThumb,
                                TargetObject = targetObject,
                                PropertyInfo = propertyInfo,
                                PointCollectionIndex = index
                            };

                            Thumb nodeThumb = new Thumb()
                            {
                                Width = THUMB_SIZE,
                                Height = THUMB_SIZE,
                                Tag = nodeThumbTarget
                            };
                            nodeThumb.DragDelta += OnNodeThumbDragDelta;

                            ToolTipService.SetToolTip(nodeThumb,
                                String.Format("{0}.{1}[{2}]", targetObject.GetType().Name,
                                                              propertyInfo.Name, 
                                                              index));

                            Point pt = ptColl[index];
                            Canvas.SetLeft(nodeThumb, pt.X - nodeThumb.Width / 2);
                            Canvas.SetTop(nodeThumb, pt.Y - nodeThumb.Height / 2);

                            characterOutlineCanvas.Children.Add(nodeThumb);
                            nodeThumbs.Add(nodeThumb);
                        }
                    }
                });

            // Set background thumb as active background
            backgroundThumb.Opacity = 0.25;
            activeBackgroundThumb = backgroundThumb;
        }

        void OnBackgroundThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            Thumb thumb = sender as Thumb;

            // Change position of thumb
            Canvas.SetLeft(thumb, Canvas.GetLeft(thumb) + args.HorizontalChange);
            Canvas.SetTop(thumb, Canvas.GetTop(thumb) + args.VerticalChange);

            // Change position of corner thumbs
            for (int corner = 0; corner < 4; corner++)
            {
                Canvas.SetLeft(cornerThumbs[corner], Canvas.GetLeft(cornerThumbs[corner]) + args.HorizontalChange);
                Canvas.SetTop(cornerThumbs[corner], Canvas.GetTop(cornerThumbs[corner]) + args.VerticalChange);
            }

            // Calculate transform for PathGeometry
            TranslateTransform xform = new TranslateTransform()
            {
                X = args.HorizontalChange,
                Y = args.VerticalChange
            };

            // Move PathGeometry
            PathGeometryHelper.Transform(thumb.Tag as PathGeometry, xform);

            // Change position of node thumbs
            foreach (Thumb nodeThumb in nodeThumbs)
            {
                Canvas.SetLeft(nodeThumb, Canvas.GetLeft(nodeThumb) + args.HorizontalChange);
                Canvas.SetTop(nodeThumb, Canvas.GetTop(nodeThumb) + args.VerticalChange);
            }
        }

        void OnCornerThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            Thumb cornerThumb = sender as Thumb;
            CornerThumbTargetInfo cornerThumbTargetInfo = cornerThumb.Tag as CornerThumbTargetInfo;
            Thumb backgroundThumb = cornerThumbTargetInfo.BackgroundThumb;

            // Watch out for negative flips!
            double horizontalChange = args.HorizontalChange;
            double verticalChange = args.VerticalChange;

            if (cornerThumbTargetInfo.Corner == Corner.UR || cornerThumbTargetInfo.Corner == Corner.LR)
                horizontalChange = Math.Max(horizontalChange, MIN_WIDTH_HEIGHT - backgroundThumb.Width);
            else
                horizontalChange = Math.Min(horizontalChange, backgroundThumb.Width - MIN_WIDTH_HEIGHT);

            if (cornerThumbTargetInfo.Corner == Corner.LL || cornerThumbTargetInfo.Corner == Corner.LR)
                verticalChange = Math.Max(verticalChange, MIN_WIDTH_HEIGHT - backgroundThumb.Height);
            else
                verticalChange = Math.Min(verticalChange, backgroundThumb.Height - MIN_WIDTH_HEIGHT);

            // Change position of thumb
            Canvas.SetLeft(cornerThumb, Canvas.GetLeft(cornerThumb) + horizontalChange);
            Canvas.SetTop(cornerThumb, Canvas.GetTop(cornerThumb) + verticalChange);

            // Create ScaleTransform for PathGeometry
            ScaleTransform xform = new ScaleTransform();

            switch (cornerThumbTargetInfo.Corner)
            {
                case Corner.UL:
                    xform.ScaleX = (backgroundThumb.Width - horizontalChange) / backgroundThumb.Width;
                    xform.ScaleY = (backgroundThumb.Height - verticalChange) / backgroundThumb.Height;
                    xform.CenterX = Canvas.GetLeft(backgroundThumb) + backgroundThumb.Width;
                    xform.CenterY = Canvas.GetTop(backgroundThumb) + backgroundThumb.Height;
                    break;

                case Corner.UR:
                    xform.ScaleX = (backgroundThumb.Width + horizontalChange) / backgroundThumb.Width;
                    xform.ScaleY = (backgroundThumb.Height - verticalChange) / backgroundThumb.Height;
                    xform.CenterX = Canvas.GetLeft(backgroundThumb);
                    xform.CenterY = Canvas.GetTop(backgroundThumb) + backgroundThumb.Height;
                    break;

                case Corner.LL:
                    xform.ScaleX = (backgroundThumb.Width - horizontalChange) / backgroundThumb.Width;
                    xform.ScaleY = (backgroundThumb.Height + verticalChange) / backgroundThumb.Height;
                    xform.CenterX = Canvas.GetLeft(backgroundThumb) + backgroundThumb.Width;
                    xform.CenterY = Canvas.GetTop(backgroundThumb);
                    break;

                case Corner.LR:
                    xform.ScaleX = (backgroundThumb.Width + horizontalChange) / backgroundThumb.Width;
                    xform.ScaleY = (backgroundThumb.Height + verticalChange) / backgroundThumb.Height;
                    xform.CenterX = Canvas.GetLeft(backgroundThumb);
                    xform.CenterY = Canvas.GetTop(backgroundThumb);
                    break;
            }

            // Transform entire PathGeometry
            PathGeometryHelper.Transform(backgroundThumb.Tag as PathGeometry, xform);

            // Move and change size of background thumb and other corner thumbs
            switch (cornerThumbTargetInfo.Corner)
            {
                case Corner.UL:
                    backgroundThumb.Width -= horizontalChange;
                    backgroundThumb.Height -= verticalChange;
                    Canvas.SetLeft(backgroundThumb, Canvas.GetLeft(backgroundThumb) + horizontalChange);
                    Canvas.SetTop(backgroundThumb, Canvas.GetTop(backgroundThumb) + verticalChange);

                    Canvas.SetLeft(cornerThumbs[(int)Corner.LL], Canvas.GetLeft(cornerThumbs[(int)Corner.LL]) + horizontalChange);
                    Canvas.SetTop(cornerThumbs[(int)Corner.UR], Canvas.GetTop(cornerThumbs[(int)Corner.UR]) + verticalChange);
                    break;

                case Corner.UR:
                    backgroundThumb.Width += horizontalChange;
                    backgroundThumb.Height -= verticalChange;
                    Canvas.SetTop(backgroundThumb, Canvas.GetTop(backgroundThumb) + verticalChange);

                    Canvas.SetLeft(cornerThumbs[(int)Corner.LR], Canvas.GetLeft(cornerThumbs[(int)Corner.LR]) + horizontalChange);
                    Canvas.SetTop(cornerThumbs[(int)Corner.UL], Canvas.GetTop(cornerThumbs[(int)Corner.UL]) + verticalChange);
                    break;

                case Corner.LL:
                    backgroundThumb.Width -= horizontalChange;
                    backgroundThumb.Height += verticalChange;
                    Canvas.SetLeft(backgroundThumb, Canvas.GetLeft(backgroundThumb) + horizontalChange);

                    Canvas.SetLeft(cornerThumbs[(int)Corner.UL], Canvas.GetLeft(cornerThumbs[(int)Corner.UL]) + horizontalChange);
                    Canvas.SetTop(cornerThumbs[(int)Corner.LR], Canvas.GetTop(cornerThumbs[(int)Corner.LR]) + verticalChange);
                    break;

                case Corner.LR:
                    backgroundThumb.Width += horizontalChange;
                    backgroundThumb.Height += verticalChange;

                    Canvas.SetLeft(cornerThumbs[(int)Corner.UR], Canvas.GetLeft(cornerThumbs[(int)Corner.UR]) + horizontalChange);
                    Canvas.SetTop(cornerThumbs[(int)Corner.LL], Canvas.GetTop(cornerThumbs[(int)Corner.LL]) + verticalChange);
                    break;
            }

            // Move all node thumbs
            foreach (Thumb nodeThumb in nodeThumbs)
            {
                Point pt = new Point(Canvas.GetLeft(nodeThumb) + nodeThumb.Width / 2, 
                                     Canvas.GetTop(nodeThumb) + nodeThumb.Height / 2);
                pt = xform.Transform(pt);
                Canvas.SetLeft(nodeThumb, pt.X - nodeThumb.Width / 2);
                Canvas.SetTop(nodeThumb, pt.Y - nodeThumb.Height / 2);
            }
        }

        void OnNodeThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            Thumb nodeThumb = sender as Thumb;

            // Change actual position of thumb
            Canvas.SetLeft(nodeThumb, Canvas.GetLeft(nodeThumb) + args.HorizontalChange);
            Canvas.SetTop(nodeThumb, Canvas.GetTop(nodeThumb) + args.VerticalChange);

            // Get the target info, the existing PathGeometry, and its Bounds
            NodeThumbTargetInfo nodeThumbTarget = nodeThumb.Tag as NodeThumbTargetInfo;
            Thumb backgroundThumb = nodeThumbTarget.BackgroundThumb;
            PathGeometry pathGeometry = backgroundThumb.Tag as PathGeometry;
            Rect boundsOld = PathGeometryHelper.CalculateBounds(pathGeometry);

            if (nodeThumbTarget.PropertyInfo.PropertyType == typeof(Point))
            {
                Point pt = (Point)nodeThumbTarget.PropertyInfo.GetValue(nodeThumbTarget.TargetObject, null);
                pt.X += args.HorizontalChange;
                pt.Y += args.VerticalChange;
                nodeThumbTarget.PropertyInfo.SetValue(nodeThumbTarget.TargetObject, pt, null);
            }
            else if (nodeThumbTarget.PropertyInfo.PropertyType == typeof(PointCollection))
            {
                PointCollection ptColl = nodeThumbTarget.PropertyInfo.GetValue(nodeThumbTarget.TargetObject, null) as PointCollection;
                Point pt = ptColl[nodeThumbTarget.PointCollectionIndex];
                pt.X += args.HorizontalChange;
                pt.Y += args.VerticalChange;
                ptColl[nodeThumbTarget.PointCollectionIndex] = pt;
            }

            Rect boundsNew = PathGeometryHelper.CalculateBounds(pathGeometry);

            // Change the position of the corner thumbs
            Canvas.SetLeft(cornerThumbs[(int)Corner.UL], Canvas.GetLeft(cornerThumbs[(int)Corner.UL]) + boundsNew.Left - boundsOld.Left);
            Canvas.SetTop(cornerThumbs[(int)Corner.UL], Canvas.GetTop(cornerThumbs[(int)Corner.UL]) + boundsNew.Top - boundsOld.Top);

            Canvas.SetLeft(cornerThumbs[(int)Corner.LL], Canvas.GetLeft(cornerThumbs[(int)Corner.UL]));
            Canvas.SetTop(cornerThumbs[(int)Corner.UR], Canvas.GetTop(cornerThumbs[(int)Corner.UL]));

            Canvas.SetLeft(cornerThumbs[(int)Corner.LR], Canvas.GetLeft(cornerThumbs[(int)Corner.UL]) + backgroundThumb.Width + boundsNew.Width - boundsOld.Width);
            Canvas.SetTop(cornerThumbs[(int)Corner.LR], Canvas.GetTop(cornerThumbs[(int)Corner.UL]) + backgroundThumb.Height + boundsNew.Height - boundsOld.Height);

            Canvas.SetLeft(cornerThumbs[(int)Corner.UR], Canvas.GetLeft(cornerThumbs[(int)Corner.LR]));
            Canvas.SetTop(cornerThumbs[(int)Corner.LL], Canvas.GetTop(cornerThumbs[(int)Corner.LR]));

            // Change the position and size of the background Thumb to reflect the new PathGeometry bounds
            Canvas.SetLeft(backgroundThumb, Canvas.GetLeft(backgroundThumb) + boundsNew.Left - boundsOld.Left);
            Canvas.SetTop(backgroundThumb, Canvas.GetTop(backgroundThumb) + boundsNew.Top - boundsOld.Top);
            backgroundThumb.Width = backgroundThumb.Width + boundsNew.Width - boundsOld.Width;
            backgroundThumb.Height = backgroundThumb.Height + boundsNew.Height - boundsOld.Height;
        }

        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            CleanUp();
        }

        void CleanUp()
        {
            // Remove all corner thumbs and node thumbs on previously active background thumb
            if (activeBackgroundThumb != null)
            {
                activeBackgroundThumb.Opacity = 0;

                for (int i = 0; i < 4; i++)
                    characterOutlineCanvas.Children.Remove(cornerThumbs[i]);

                foreach (Thumb nodeThumb in nodeThumbs)
                    characterOutlineCanvas.Children.Remove(nodeThumb);
            }
            activeBackgroundThumb = null;
        }
    }

    enum Corner
    {
        UL, UR, LL, LR
    };

    class CornerThumbTargetInfo
    {
        public Thumb BackgroundThumb { set; get; }
        public Corner Corner { set; get; }
    }

    class NodeThumbTargetInfo
    {
        public Thumb BackgroundThumb { set; get; }
        public object TargetObject { set; get; }
        public PropertyInfo PropertyInfo { set; get; }
        public int PointCollectionIndex { set; get; }
    }
}
