﻿// PathGeometryHelper.cs (c) 2009 by Charles Petzold
using System;
using System.Reflection;
using System.Windows;
using System.Windows.Media;

namespace ManipulateCharacterOutlines
{
    public delegate void PathGeometryEnumerateCallback(object obj, PropertyInfo propInfo);

    public static class PathGeometryHelper
    {
        public static void Transform(PathGeometry pathGeometry, Transform xform)
        {
            foreach (PathFigure pathFigure in pathGeometry.Figures)
            {
                pathFigure.StartPoint = xform.Transform(pathFigure.StartPoint);

                foreach (PathSegment pathSegment in pathFigure.Segments)
                {
                    if (pathSegment is LineSegment)
                    {
                        LineSegment lineSegment = pathSegment as LineSegment;
                        lineSegment.Point = xform.Transform(lineSegment.Point);
                    }
                    else if (pathSegment is PolyLineSegment)
                    {
                        PolyLineSegment polyLineSegment = pathSegment as PolyLineSegment;

                        for (int i = 0; i < polyLineSegment.Points.Count; i++)
                            polyLineSegment.Points[i] = xform.Transform(polyLineSegment.Points[i]);
                    }
                    else if (pathSegment is BezierSegment)
                    {
                        BezierSegment bezierSegment = pathSegment as BezierSegment;
                        bezierSegment.Point1 = xform.Transform(bezierSegment.Point1);
                        bezierSegment.Point2 = xform.Transform(bezierSegment.Point2);
                        bezierSegment.Point3 = xform.Transform(bezierSegment.Point3);
                    }
                    else if (pathSegment is PolyBezierSegment)
                    {
                        PolyBezierSegment polyBezierSegment = pathSegment as PolyBezierSegment;

                        for (int i = 0; i < polyBezierSegment.Points.Count; i++)
                            polyBezierSegment.Points[i] = xform.Transform(polyBezierSegment.Points[i]);
                    }
                    else
                        throw new NotImplementedException("PathGeometryHelper.Transform does not yet support PathSegment " +
                                                          "objects of type " + pathSegment.GetType().Name);
                }
            }
        }
        public static void Enumerate(PathGeometry pathGeometry, PathGeometryEnumerateCallback callback)
        {
            foreach (PathFigure pathFigure in pathGeometry.Figures)
            {
                callback(pathFigure, pathFigure.GetType().GetProperty("StartPoint"));

                foreach (PathSegment pathSegment in pathFigure.Segments)
                {
                    if (pathSegment is LineSegment)
                    {
                        callback(pathSegment, pathSegment.GetType().GetProperty("Point"));
                    }
                    else if (pathSegment is PolyLineSegment || pathSegment is PolyBezierSegment)
                    {
                        callback(pathSegment, pathSegment.GetType().GetProperty("Points"));
                    }
                    else if (pathSegment is BezierSegment)
                    {
                        callback(pathSegment, pathSegment.GetType().GetProperty("Point1"));
                        callback(pathSegment, pathSegment.GetType().GetProperty("Point2"));
                        callback(pathSegment, pathSegment.GetType().GetProperty("Point3"));
                    }
                    else
                        throw new NotImplementedException("PathGeometryHelper.Enumerate does not yet support PathSegment " +
                                                          "objects of type " + pathSegment.GetType().Name);
                }
            }
        }

        // Added because Bounds wasn't returning new values
        public static Rect CalculateBounds(PathGeometry pathGeometry)
        {
            double xMin = Double.PositiveInfinity;
            double xMax = Double.NegativeInfinity;
            double yMin = Double.PositiveInfinity;
            double yMax = Double.NegativeInfinity;

            foreach (PathFigure pathFigure in pathGeometry.Figures)
            {
                NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, pathFigure.StartPoint);

                foreach (PathSegment pathSegment in pathFigure.Segments)
                {
                    if (pathSegment is LineSegment)
                    {
                        LineSegment lineSegment = pathSegment as LineSegment;
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, lineSegment.Point);
                    }
                    else if (pathSegment is PolyLineSegment)
                    {
                        PolyLineSegment polyLineSegment = pathSegment as PolyLineSegment;

                        for (int i = 0; i < polyLineSegment.Points.Count; i++)
                            NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, polyLineSegment.Points[i]);
                    }
                    else if (pathSegment is BezierSegment)
                    {
                        BezierSegment bezierSegment = pathSegment as BezierSegment;
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, bezierSegment.Point1);
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, bezierSegment.Point2);
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, bezierSegment.Point3);
                    }
                    else if (pathSegment is PolyBezierSegment)
                    {
                        PolyBezierSegment polyBezierSegment = pathSegment as PolyBezierSegment;

                        for (int i = 0; i < polyBezierSegment.Points.Count; i++)
                            NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, polyBezierSegment.Points[i]);
                    }
                    else
                        throw new NotImplementedException("PathGeometryHelper.Transform does not yet support PathSegment " +
                                                          "objects of type " + pathSegment.GetType().Name);
                }
            }

            return new Rect(new Point(xMin, yMin), new Point(xMax, yMax));
        }

        static void NewMinMax(ref double xMin, ref double xMax, ref double yMin, ref double yMax, Point pt)
        {
            xMin = Math.Min(xMin, pt.X);
            xMax = Math.Max(xMax, pt.X);
            yMin = Math.Min(yMin, pt.Y);
            yMax = Math.Max(yMax, pt.Y);
        }
    }
}
